/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_UBLASUTILS_ORTHONORMALIZE_HPP
#define MYFITTER_UBLASUTILS_ORTHONORMALIZE_HPP

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <iostream>

namespace myfitter { namespace ublasutils { 

    template<class realT, class L, class A>
    typename boost::numeric::ublas::matrix<std::complex<realT>, L, A>::size_type
    orthonormalize(boost::numeric::ublas::matrix<std::complex<realT>, L, A>& M,
                   realT tolerance,
                   typename boost::numeric::ublas::
                   matrix<std::complex<realT>, L, A>::size_type start=0,
                   int stop=-1)
    {
        using boost::numeric::ublas::matrix;
        using boost::numeric::ublas::matrix_range;
        using boost::numeric::ublas::matrix_row;
        using boost::numeric::ublas::matrix_column;
        using boost::numeric::ublas::vector;
        using boost::numeric::ublas::vector_range;

        typedef std::complex<realT> complT;

        typedef matrix<complT, L, A> compl_mat;
        typedef vector<complT, A> compl_vec;

        typedef typename compl_mat::size_type size_type;

        typedef matrix_range<compl_mat> compl_mat_range;
        typedef matrix_column<compl_mat> compl_mat_col;
        typedef matrix_row<compl_mat> compl_mat_row;
        typedef vector_range<compl_mat_col> compl_mat_col_range;
        typedef vector_range<compl_mat_row> compl_mat_row_range;

        size_type n = M.size2();
        if(stop >= 0) n = stop;
        size_type k = start;
        size_type i;
        for(i=start; i<n; ++i) {
            compl_mat_col vi(M, i);
            compl_mat_col uk(M, k);
            if(k<i) uk = vi;
            for(size_type j=0; j<k; ++j) {
                compl_mat_col uj(M,j);
                uk -= inner_prod(conj(uj), uk)*uj;
            }
            realT r = std::sqrt(std::real(inner_prod(conj(uk), uk)));
            if(r < tolerance) continue;
            uk *= 1./r;
            ++k;
            if(k > M.size1()) { ++i; break; }
        }
        return k;
    }
            
    template<class realT, class L, class A>
    typename boost::numeric::ublas::matrix<realT, L, A>::size_type
    orthonormalize(boost::numeric::ublas::matrix<realT, L, A>& M,
                   realT tolerance,
                   typename boost::numeric::ublas::
                   matrix<realT, L, A>::size_type
                   start=0, int stop=-1)
    {
        using boost::numeric::ublas::matrix;
        using boost::numeric::ublas::matrix_range;
        using boost::numeric::ublas::matrix_row;
        using boost::numeric::ublas::matrix_column;
        using boost::numeric::ublas::vector;
        using boost::numeric::ublas::vector_range;

        typedef matrix<realT, L, A> real_mat;
        typedef vector<realT, A> real_vec;

        typedef typename real_mat::size_type size_type;
        typedef matrix_column<real_mat> real_mat_col;

        size_type n = M.size2();
        if(stop >= 0) n = stop;
        size_type k = start;
        size_type i;
        for(i=start; i<n; ++i) {
            real_mat_col vi(M, i);
            real_mat_col uk(M, k);
            if(k<i) uk = vi;
            for(size_type j=0; j<k; ++j) {
                real_mat_col uj(M,j);
                uk -= inner_prod(conj(uj), uk)*uj;
            }
            realT r = std::sqrt(inner_prod(uk, uk));
            if(r < tolerance) continue;
            uk *= 1./r;
            ++k;
            if(k > M.size1()) { ++i; break; }
        }
        return k;
    }
            
}}


#endif


